@@ -3,8 +3,13 @@
 # Accelerator for pip, the Python package manager.
 #
 # Author: Peter Odding <peter.odding@paylogic.eu>
-# Last Change: April 17, 2013
+# Last Change: April 23, 2013
 # URL: https://github.com/paylogic/pip-accel
+#
+# TODO Consider using pip's API instead of running it as a subprocess.
+# TODO Permanently store logs in the pip-accel directory (think about log rotation).
+# TODO Maybe we should save the output of `python setup.py bdist_dumb` somewhere as well?
+# FIXME When we run pip to download source distributions it's silent for a long time... (buffered output)
 
 """
 Usage: pip-accel [ARGUMENTS TO PIP]
@@ -23,7 +28,6 @@ import os
 import os.path
 import pkg_resources
 import re
-import shutil
 import sys
 import tarfile
 import time
@@ -36,28 +40,42 @@ INTERACTIVE = os.isatty(1)
 # Check if the operator requested verbose output.
 VERBOSE = '-v' in sys.argv
 
+# Find the environment where requirements are to be installed.
+ENVIRONMENT = os.path.abspath(os.environ.get('VIRTUAL_ENV', sys.prefix))
+
 # The main loop of pip-accel retries at most this many times to counter pip errors
 # due to connectivity issues with PyPi and/or linked distribution websites.
 MAX_RETRIES = 10
 
+# The directories under a given environment (/usr, /usr/local or a virtual
+# environment) where we can install files from binary distribution archives.
+KNOWN_ROOTS = ('bin/', 'lib/', 'man/', 'share/')
+
+# The version number of the binary distribution cache format in use. When we
+# break backwards compatibility we bump this number so that pip-accel knows it
+# should clear the cache before proceeding.
+CACHE_FORMAT_REVISION = 2
+
 # Select the default location of the download cache and other files based on
 # the user running the pip-accel command (root goes to /var/cache/pip-accel,
 # otherwise ~/.pip-accel).
 if os.getuid() == 0:
     download_cache = '/root/.pip/download-cache'
-    source_index = '/var/cache/pip-accel/sources'
-    binary_index = '/var/cache/pip-accel/binaries'
+    pip_accel_cache = '/var/cache/pip-accel'
 else:
     download_cache = os.path.expanduser('~/.pip/download-cache')
-    source_index = os.path.expanduser('~/.pip-accel/sources')
-    binary_index = os.path.expanduser('~/.pip-accel/binaries')
+    pip_accel_cache = os.path.expanduser('~/.pip-accel')
 
 # Enable overriding the default locations with environment variables.
 if 'PIP_DOWNLOAD_CACHE' in os.environ:
     download_cache = os.path.expanduser(os.environ['PIP_DOWNLOAD_CACHE'])
 if 'PIP_ACCEL_CACHE' in os.environ:
-    source_index = os.path.join(os.path.expanduser(os.environ['PIP_ACCEL_CACHE']), 'sources')
-    binary_index = os.path.join(os.path.expanduser(os.environ['PIP_ACCEL_CACHE']), 'binaries')
+    pip_accel_cache = os.path.expanduser(os.environ['PIP_ACCEL_CACHE'])
+
+# Generate the absolute pathnames of the source/binary caches.
+source_index = os.path.join(pip_accel_cache, 'sources')
+binary_index = os.path.join(pip_accel_cache, 'binaries')
+index_version_file = os.path.join(pip_accel_cache, 'version.txt')
 
 def main():
     """
@@ -75,18 +93,24 @@ def main():
     initialize_directories()
     # Execute "pip install" in a loop in order to retry after intermittent
     # error responses from servers (which can happen quite frequently).
-    for i in xrange(1, MAX_RETRIES):
-        requirements = unpack_source_dists(arguments)
-        if requirements is None:
-            download_source_dists(arguments)
-        elif not requirements:
-            message("No requirements found in pip's output, probably there's nothing to do.\n")
-            return
-        else:
-            if build_binary_dists(requirements) and install_requirements(requirements):
-                message("Done! Took %s to install %i package%s.\n", main_timer, len(requirements), '' if len(requirements) == 1 else 's')
-            return
-    message("External command failed %i times, aborting!\n" % MAX_RETRIES)
+    try:
+        for i in xrange(1, MAX_RETRIES):
+            requirements = unpack_source_dists(arguments)
+            if requirements is None:
+                download_source_dists(arguments)
+            elif not requirements:
+                message("No requirements found in pip's output, probably there's nothing to do.\n")
+                return
+            else:
+                if build_binary_dists(requirements) and install_requirements(requirements):
+                    message("Done! Took %s to install %i package%s.\n", main_timer, len(requirements), '' if len(requirements) == 1 else 's')
+                return
+    except InstallationError:
+        # Abort early when pip reports installation errors.
+        message("Error: Pip reported unrecoverable installation errors. Please fix and rerun!\n")
+        sys.exit(1)
+    # Abort when after N retries we still failed to download source distributions.
+    message("Error: External command failed %i times, aborting!\n" % MAX_RETRIES)
     sys.exit(1)
 
 def unpack_source_dists(arguments):
@@ -221,7 +245,7 @@ def build_binary_dists(requirements):
         # Build a binary distribution.
         os.chdir(directory)
         message("Building binary distribution of %s (%s) ..\n", name, version)
-        status = (os.system('python setup.py bdist') == 0)
+        status = (os.system('python setup.py bdist_dumb --format=gztar') == 0)
         os.chdir(old_directory)
         if not status:
             message("Failed to build binary distribution!\n")
@@ -233,12 +257,52 @@ def build_binary_dists(requirements):
             return False
         cache_file = '%s:%s.tar.gz' % (name, version)
         message("Copying binary distribution %s to cache as %s.\n", filenames[0], cache_file)
-        shutil.move(os.path.join(directory, 'dist', filenames[0]),
-                    os.path.join(binary_index, cache_file))
+        cache_binary_distribution(os.path.join(directory, 'dist', filenames[0]),
+                                  os.path.join(binary_index, cache_file))
     message("Finished building binary distributions.\n")
     return True
 
-def install_requirements(requirements, install_prefix=sys.prefix):
+def cache_binary_distribution(input_path, output_path):
+    """
+    Transform a binary distribution archive created with `python setup.py
+    bdist_dumb --format=gztar` into a form that can be cached for future use.
+    This comes down to making the pathnames inside the archive relative to the
+    "prefix" that the binary distribution was built for.
+
+    Expects two arguments: The pathname of the original binary distribution
+    archive and the pathname of the binary distribution in the cache
+    directory.
+    """
+    # Copy the tar archive file by file so we can rewrite their pathnames.
+    debug("Expected prefix in binary distribution archive: %s\n", ENVIRONMENT)
+    input_bdist = tarfile.open(input_path, 'r:gz')
+    output_bdist = tarfile.open(output_path, 'w:gz')
+    for member in input_bdist.getmembers():
+        # In my testing the `dumb' tar files created with the `python setup.py
+        # bdist' command contain pathnames that are relative instead of
+        # absolute, but they're relative to / which is kind of awkward: I would
+        # like to use os.path.relpath() on them but that won't give the correct
+        # result without preprocessing...
+        original_pathname = member.name
+        absolute_pathname = re.sub(r'^\./', '/', original_pathname)
+        if member.isdir() or member.isdev():
+            debug("Ignoring directory or device file: %s\n", absolute_pathname)
+        else:
+            modified_pathname = os.path.relpath(absolute_pathname, ENVIRONMENT)
+            debug("Transformed %s -> %s.\n", original_pathname, modified_pathname)
+            if modified_pathname.startswith(KNOWN_ROOTS):
+                # Get the file data from the input archive.
+                file_data = input_bdist.extractfile(original_pathname)
+                # Use all metadata from the input archive but modify the filename.
+                member.name = modified_pathname
+                # Copy modified metadata + original file data to output archive.
+                output_bdist.addfile(member, file_data)
+            else:
+                message("Warning: Ignoring file in unknown location: %s\n", original_pathname)
+    input_bdist.close()
+    output_bdist.close()
+
+def install_requirements(requirements, install_prefix=ENVIRONMENT):
     """
     Manually install all requirements from binary distributions.
 
@@ -260,7 +324,7 @@ def install_requirements(requirements, install_prefix=sys.prefix):
     message("Finished installing all requirements in %s.\n", install_timer)
     return True
 
-def install_binary_dist(filename, install_prefix=sys.prefix):
+def install_binary_dist(filename, install_prefix=ENVIRONMENT):
     """
     Install a binary distribution created with `python setup.py bdist` into the
     given prefix (a directory like /usr, /usr/local or a virtual environment).
@@ -276,53 +340,24 @@ def install_binary_dist(filename, install_prefix=sys.prefix):
     python = os.path.join(install_prefix, 'bin', 'python')
     message("Installing binary distribution %s to %s ..\n", filename, install_prefix)
     archive = tarfile.open(filename, 'r:gz')
-    for original_path, relative_path, mode in find_bdist_contents(archive):
-        install_path = os.path.join(install_prefix, relative_path)
+    for member in archive.getmembers():
+        install_path = os.path.join(install_prefix, member.name)
         directory = os.path.dirname(install_path)
         if not os.path.isdir(directory):
             os.makedirs(directory)
         debug("Writing %s\n", install_path)
-        file_handle = archive.extractfile(original_path)
+        file_handle = archive.extractfile(member)
         with open(install_path, 'w') as handle:
             contents = file_handle.read()
             if contents.startswith('#!/'):
                 # Fix hashbangs.
+                # TODO Don't be so brute force about it?
                 contents = re.sub('^#![^\n]+', '#!' + python, contents)
             handle.write(contents)
-        os.chmod(install_path, mode)
+        os.chmod(install_path, member.mode)
     archive.close()
     message("Finished installing binary distribution in %s.\n", install_timer)
 
-def find_bdist_contents(archive):
-    """
-    Transform the absolute pathnames embedded in a binary distribution into
-    relative filenames that can be prefixed by /usr, /usr/local or the path to
-    a virtual environment.
-
-    Expects one argument: a tarfile object.
-
-    Returns a list of tuples with three values each: (original-path,
-    relative-path, file-mode). The first value is the pathname from the tar
-    archive, the second value is the transformed pathname and the third value
-    contains the integer mode that the file should get (executable bits and
-    other file permissions).
-    """
-    contents = []
-    for member in archive.getmembers():
-        original_path = member.name
-        member_is_file = member.isfile()
-        for substring in ['bin', 'lib', 'man', 'share']:
-            tokens = original_path.split('/%s/' % substring, 1)
-            if len(tokens) >= 2:
-                relative_path = os.path.join(substring, tokens[1])
-                if relative_path != original_path and member_is_file:
-                    contents.append((original_path, relative_path, member.mode))
-                    break  # the inner for loop (note also the else below)
-        else:
-            if member_is_file:
-                message("Warning: Ignoring unmatched file %s\n", original_path)
-    return contents
-
 def run_pip(arguments, use_remote_index):
     """
     Execute a modified `pip install` command.
@@ -333,6 +368,10 @@ def run_pip(arguments, use_remote_index):
 
     Returns a list of strings with the lines of output from `pip` on success,
     None otherwise (`pip` exited with a nonzero exit status).
+
+    Note that if pip reports an unrecoverable installation error, the custom
+    exception type `InstallationError` is raised so that the pip accelerator
+    can abort immediately instead of entering the retry loop.
     """
     command_line = []
     for i, arg in enumerate(arguments):
@@ -352,6 +391,8 @@ def run_pip(arguments, use_remote_index):
     for line in pip:
         message("  %s\n", line.rstrip(), prefix=False)
         output.append(line)
+        if re.search(r'\bInstallationError:', line):
+            raise InstallationError
     if pip.close() is None:
         update_source_dists_index()
         return output
@@ -418,12 +459,27 @@ def add_extension(download_path, archive_path):
 def initialize_directories():
     """
     Create the directories for the download cache, the source index and the
-    binary index if any of them don't exist yet.
+    binary index if any of them don't exist yet and reset the binary index
+    when its format changes.
     """
     # Create all required directories on the fly.
     for directory in [download_cache, source_index, binary_index]:
         if not os.path.isdir(directory):
             os.makedirs(directory)
+    # Invalidate the binary distribution cache when the
+    # format is changed in backwards incompatible ways.
+    if os.path.isfile(index_version_file):
+        with open(index_version_file) as handle:
+            if int(handle.read()) == CACHE_FORMAT_REVISION:
+                debug("Binary distribution cache format is compatible.\n")
+                return
+    debug("Binary distribution cache format is incompatible; clearing cache ..\n")
+    for entry in os.listdir(binary_index):
+        pathname = os.path.join(binary_index, entry)
+        debug(" - Deleting %s\n", pathname)
+        os.unlink(pathname)
+    with open(index_version_file, 'w') as handle:
+        handle.write("%i\n" % CACHE_FORMAT_REVISION)
 
 def interactive_message(text):
     """
@@ -456,12 +512,21 @@ def message(text, *args, **kw):
     the `%` operator. The prefix can be disabled by passing the keyword
     argument `prefix=False`.
     """
+    # TODO Maybe cleaner to use Python's logging module with a custom formatter?
     if kw.get('prefix', True):
         text = '(pip-accel) ' + text
     if INTERACTIVE:
         text = '\r' + text
     sys.stderr.write(text % args)
 
+class InstallationError(Exception):
+    """
+    Custom exception type used to signal that pip reported an unrecoverable
+    installation error. The pip accelerator can't do anything useful when pip
+    reports such errors, so it will simply abort.
+    """
+    pass
+
 class Timer:
 
     """
